module.exports = (logger) => {
  'use strict';

  const log = logger('fs-oauth-handler'),
    axios = require('axios'),
    crypto = require('crypto');

  const hasProp = (o, ...pp) => {
    if (!pp.length) {
      return false;
    }

    for (const p of pp) {
      if (typeof o === 'object' && Object.prototype.hasOwnProperty.call(o, p)) {
        o = o[p];
      } else {
        return false;
      }
    }
    return true;
  };

  const fsOauthHandler = (axios, crypto, log) => {
    class Token {
      constructor(tokenResponse) {
        if (
          !hasProp(tokenResponse, 'token') ||
          !(Object(tokenResponse['token']) instanceof String) ||
          // Check if the token looks like a valid JWT token
          !tokenResponse['token'].match(
            /^[a-zA-Z0-9_-]+\.[a-zA-Z0-9_-]+(?:\.[a-zA-Z0-9_-]+)?$/
          )
        ) {
          throw new Error('Invalid or missing token information');
        }

        // Retrieve the token information
        const tokenContent = tokenResponse.token.split('.', 2)[1];

        // Attempt to decode as base64(url)
        const tokenDecodedContent = Buffer.from(
          tokenContent,
          'base64'
        ).toString('utf-8');

        // Now, parse the JSON inside. This is needed to get the expiry time
        const tokenDecodedJsonContent = JSON.parse(tokenDecodedContent);

        log.debug('Token content', tokenDecodedContent);

        // Get the expiry time. If it is not found, assume that the token is valid
        // for 2 minutes (120s)
        const expiryTime = hasProp(tokenDecodedJsonContent, 'exp')
          ? tokenDecodedJsonContent['exp'] * 1e3
          : Date.now() + 120 * 1e3;

        this.expiration = expiryTime;
        this.accessToken = tokenResponse['token'];
        this.tokenType = 'bearer';
      }

      get isFresh() {
        return Date.now() < this.expiration;
      }

      expiresWithin(ms) {
        return this.expiration - Date.now() < ms;
      }

      get authorizationString() {
        switch (this.tokenType) {
          case 'bearer':
            return 'Bearer ' + this.accessToken;
          default:
            throw new Error('Token type not implemented');
        }
      }
    }

    const axios_ = axios.create(),
      tokenMap = new Map(),
      tokenRefreshMap = new Map();

    axios_.defaults.headers.common['Accept'] = 'application/json;version=1';
    axios_.defaults.headers.common['Accept-Charset'] = 'UTF-8';
    axios_.defaults.headers.post['Content-Type'] = 'application/json;version=1';

    /* Requests a new OAuth token from FS */
    const requestToken = (config) => {
      const tokenPath = config['fsOauth']['tokenPath'] || '/token';
      return axios_
        .get(
          tokenPath,
          Object.assign({}, config, {
            auth: {
              username: config['fsOauth']['username'],
              password: config['fsOauth']['password'],
            },
            fsOauth: undefined,
            params: undefined,
            transformRequest: undefined,
            transformResponse: undefined,
            paramsSerializer: null,
            responseType: 'json',
            responseEncoding: 'utf8',
            maxContentLength: 4096,
            validateStatus: (status) => status === 200,
          })
        )
        .then((resp) => {
          if (hasProp(resp, 'data')) {
            const token = new Token(resp['data']);
            return Promise.resolve(Object.freeze(token));
          }
        });
    };

    const getMapKey = (config) => {
      const applicationKey = config['fsOauth']['username'];
      const secret = config['fsOauth']['password'];

      /* Generate a map key based on the current credentials */
      const key = crypto
        .createHmac('sha256', secret)
        .update(applicationKey)
        .update('@')
        .update(config['baseURL'] || '\0')
        .update(config['fsOauth']['tokenPath'] || '\0')
        .digest('hex');

      return key;
    };

    /* Obtain a token based on the current credentials */
    const obtainToken = (config, forceRefresh) => {
      if (
        !hasProp(config, 'fsOauth', 'username') ||
        !hasProp(config['fsOauth'], 'password')
      ) {
        log.debug('No username or password. Passing through.');
        return null;
      }

      const key = getMapKey(config),
        logKey = `[${key.substr(0, 8)}]`;

      /* If there already is a token, attempt to use it */
      if (!forceRefresh && tokenMap.has(key)) {
        return new Promise((resolve, reject) => {
          const tokenPromise = tokenMap.get(key);
          tokenPromise
            .then((token) => {
              /* Is the current token valid? */
              if (token.isFresh) {
                log.debug('Fresh token found', logKey);
                resolve(token);
                /* If the current token is still valid, but about to expire,
                 * proactively request a new one. TODO: The threshold is hardcoded
                 * to 100s, but having something configurable would be best. */
                if (token.expiresWithin(1e5) && !tokenRefreshMap.has(key)) {
                  log.debug('Proactively requesting a new token', logKey);
                  const tokenRefreshPromise = requestToken(config);
                  tokenRefreshMap.set(key, tokenRefreshPromise);
                  tokenRefreshPromise.then((token) => {
                    tokenMap.set(key, Promise.resolve(token));
                    tokenRefreshMap.delete(key);
                  });
                }
              } else {
                /* The token has already expired, so we need to obtain a new one */
                log.debug(
                  'Token has expired. Another one will be requested',
                  logKey
                );
                resolve(obtainToken(config, true));
              }
            })
            .catch(reject);
        });
      }

      /* No token found, so request a new one */
      if (!forceRefresh) {
        log.debug('No token found, requesting one', logKey);
      } else if (!tokenMap.has(key)) {
        log.debug('No token found, requesting one (forced)', logKey);
      } else {
        log.debug(
          'A token has been found, but still requesting one (forced)',
          logKey
        );
      }
      const tokenPromise = requestToken(config);
      tokenMap.set(key, tokenPromise);
      return tokenPromise;
    };

    /* This interceptor takes care of adding a FS token to requests */
    axios_.interceptors.request.use((config) => {
      if (!hasProp(config, 'fsOauth') || !config['fsOauth']) {
        return config;
      }

      const tokenPromise = obtainToken(config);

      return tokenPromise.then((token) => {
        if (!(token instanceof Token)) {
          return config;
        }

        config['headers']['common']['Authorization'] =
          token.authorizationString;
        return config;
      });
    });

    /* The token used could have expired. In this case, obtain a new token and re-attempt the
     * request */
    axios_.interceptors.response.use(
      (resp) => resp,
      (error) => {
        if (
          hasProp(error, 'response', 'data', 'errors') &&
          hasProp(error['response'], 'status') &&
          Array.isArray(error['response']['data']['errors']) &&
          error['response']['data']['errors'].length === 1 &&
          hasProp(error, 'config', 'fsOauth') &&
          error['config']['fsOauth']
        ) {
          if (
            error['response']['status'] === 500 &&
            hasProp(error['response']['data']['errors'][0], 'code') &&
            error['response']['data']['errors'][0]['code'] === 50013 &&
            (!hasProp(error['config']['fsOauth'], 'refreshToken') ||
              error['config']['fsOauth']['refreshToken'])
          ) {
            log.debug(error['response']['data'], 'Token has expired');
            /* const tokenPromise = */ obtainToken(error['config'], true);
            const config = error['config'];
            Object.assign(config['fsOauth'], {refreshToken: false});
            return Promise.resolve(axios_.request(config));
          }
        }
        return Promise.reject(error);
      }
    );

    return axios_;
  };

  return fsOauthHandler(axios, crypto, log);
};
